#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Serialization Tools. """

import uuid
import warnings
import inspect
import collections
import collections.abc
from ast import literal_eval
from typing import get_origin, get_args, Union, Any, BinaryIO, TextIO
from numpy import int64, float64
import networkx as nx
from dessia_common import REF_MARKER, OLD_REF_MARKER
import dessia_common.errors as dc_err
from dessia_common.files import StringFile, BinaryFile
import dessia_common.utils.types as dcty
from dessia_common.utils.helpers import full_classname, get_python_class_from_class_name
from dessia_common.abstract import CoreDessiaObject
from dessia_common.typings import InstanceOf, JsonSerializable
from dessia_common.measures import Measure
from dessia_common.graph import explore_tree_from_leaves
from dessia_common.breakdown import get_in_object_from_path, set_in_object_from_path
from dessia_common.schemas.core import TYPING_EQUIVALENCES, is_typing, serialize_annotation

fullargsspec_cache = {}


class SerializableObject(CoreDessiaObject):
    """ Serialization capabilities of Dessia Object. """

    _standalone_in_db = False
    _non_serializable_attributes = []

    def base_dict(self):
        """ A base dict for to_dict: set up a dict with object class and version. """
        dict_ = {'object_class': self.full_classname}
        return dict_

    def _serializable_dict(self):
        """
        Return a dict of attribute_name, values (still python, not serialized).

        Keys are filtered with non serializable attributes controls.
        """
        dict_ = {k: v for k, v in self.__dict__.items()
                 if k not in self._non_serializable_attributes and not k.startswith('_')}
        return dict_

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#',
                id_method=True, id_memo=None, **kwargs) -> JsonSerializable:
        """ Generic to_dict method. """
        if memo is None:
            memo = {}

        # Default to dict
        serialized_dict = self.base_dict()
        dict_ = self._serializable_dict()
        if use_pointers:
            serialized_dict.update(serialize_dict_with_pointers(dict_, memo=memo, path=path, id_method=id_method,
                                                                id_memo=id_memo)[0])
        else:
            serialized_dict.update(serialize_dict(dict_))

        return serialized_dict

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, **kwargs) -> 'SerializableObject':
        """ Generic dict_to_object method. """
        if 'object_class' in dict_:
            return dict_to_object(dict_=dict_, **kwargs)
        if cls is not SerializableObject:
            return dict_to_object(dict_=dict_, class_=cls, **kwargs)
        raise NotImplementedError("No object_class in dict")

    @property
    def full_classname(self):
        """ Full class name of class like: package.module.submodule.classname. """
        return full_classname(self)


def serialize_dict(dict_):
    """ Serialize dictionary values. """
    return {k: serialize(v) for k, v in dict_.items()}


def serialize_sequence(seq):
    """ Serialize a sequence (list or sequence) into a list of dictionaries. """
    return [serialize(v) for v in seq]


def serialize(value):
    """
    Main function for serialization without pointers.

    Calls recursively itself serialize_sequence and serialize_dict.
    """
    if isinstance(value, SerializableObject):
        try:
            serialized_value = value.to_dict(use_pointers=False)
        except TypeError:
            warnings.warn(f'specific to_dict of class {value.__class__} '
                          'should implement use_pointers, memo and path arguments', Warning)
            serialized_value = value.to_dict()
    elif isinstance(value, dict):
        serialized_value = serialize_dict(value)
    elif dcty.is_sequence(value):
        serialized_value = serialize_sequence(value)
    elif isinstance(value, (BinaryFile, StringFile)):
        serialized_value = value
    elif isinstance(value, int64):
        serialized_value = int(value)
    elif isinstance(value, float64):
        serialized_value = float(value)
    elif isinstance(value, type) or is_typing(value):
        return serialize_annotation(value)
    elif hasattr(value, 'to_dict'):
        to_dict_method = getattr(value, 'to_dict', None)
        if callable(to_dict_method):
            return to_dict_method()
    else:
        if not dcty.is_jsonable(value):
            raise dc_err.SerializationError(f"Element of value '{value}' is not json serializable")
        serialized_value = value
    return serialized_value


def serialize_with_pointers(value, memo=None, path='#', id_method=True, id_memo=None):
    """ Main function for serialization with pointers. """
    if memo is None:
        memo = {}
    if id_memo is None:
        id_memo = {}

    if isinstance(value, SerializableObject):
        if value in memo:
            path_to_refs, serialized_value, id_, _ = memo[value]
            id_memo[id_] = serialized_value
            return {REF_MARKER: path_to_refs}, memo
        try:
            serialized = value.to_dict(use_pointers=True, memo=memo, path=path, id_memo=id_memo)
        except TypeError:
            warnings.warn('specific to_dict should implement use_pointers, memo, path and id_memo arguments', Warning)
            serialized = value.to_dict()

        if id_method:
            id_ = str(uuid.uuid1())
            path_to_refs = f"#/_references/{id_}"
            memo[value] = path_to_refs, serialized, id_, path
            if value._standalone_in_db:
                id_memo[id_] = serialized
                serialized = {REF_MARKER: path_to_refs}
        else:
            memo[value] = path, serialized, None, path

    elif isinstance(value, type):
        # TODO Why do we serialize types with pointers ? These are only just strings.
        if value in memo:
            path_to_refs, serialized_value, id_, _ = memo[value]
            id_memo[id_] = serialized_value
            return {REF_MARKER: memo[value]}, memo
        serialized = serialize_annotation(value)

    # Regular object
    elif hasattr(value, 'to_dict'):
        if value in memo:
            path_to_refs, serialized_value, id_, path_to_value = memo[value]
            id_memo[id_] = serialized_value
            return {REF_MARKER: path_to_value}, memo
        serialized = value.to_dict()

        if id_method:
            id_ = str(uuid.uuid1())
            path_to_refs = f"#/_references/{id_}"
            memo[value] = path_to_refs, serialized, id_, path
        else:
            memo[value] = path, serialized, None, path

    elif isinstance(value, dict):
        serialized, memo = serialize_dict_with_pointers(value, memo=memo, path=path, id_method=id_method,
                                                        id_memo=id_memo)
    elif dcty.is_sequence(value):
        serialized, memo = serialize_sequence_with_pointers(value, memo=memo, path=path, id_method=id_method,
                                                            id_memo=id_memo)

    elif isinstance(value, (BinaryFile, StringFile)):
        serialized = value
    elif isinstance(value, (int64, float64)):
        serialized = serialize(value)
    else:
        if not dcty.is_jsonable(value):
            msg = f'Element of value {value} (type: {value.__class__.__name__}) is not json serializable'
            raise dc_err.SerializationError(msg)
        serialized = value

    if path == '#' and id_method:
        add_references(serialized, memo, id_memo)
    return serialized, memo


def serialize_dict_with_pointers(dict_, memo, path, id_method, id_memo):
    """
    Recursively serialize a dictionary with JSON pointers.

    Use a memo dict at a given path of the top level object.
    """
    if memo is None:
        memo = {}
    if id_memo is None:
        id_memo = {}

    serialized_dict = {}
    dict_attrs_keys = []
    seq_attrs_keys = []
    other_keys = []
    # Detecting type of keys
    for key, value in dict_.items():
        if isinstance(value, dict):
            dict_attrs_keys.append(key)
        elif dcty.is_sequence(value):
            seq_attrs_keys.append(key)
        else:
            other_keys.append(key)

    for key in other_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_with_pointers(dict_[key], memo=memo, path=value_path,
                                                             id_memo=id_memo)
    # Handle sequence & dictionaries afterwards
    for key in seq_attrs_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_sequence_with_pointers(dict_[key], memo=memo, path=value_path,
                                                                      id_method=id_method, id_memo=id_memo)

    for key in dict_attrs_keys:
        value_path = f'{path}/{key}'
        serialized_dict[key], memo = serialize_dict_with_pointers(dict_[key], memo=memo, path=value_path,
                                                                  id_method=id_method, id_memo=id_memo)

    if path == '#':
        add_references(serialized_dict, memo, id_memo)
    return serialized_dict, memo


def add_references(dict_, memo, id_memo):
    """ Add _references to a dict given the memos. """
    dict_['_references'] = id_memo

    # Rewriting dc__refs
    for _, serialized, id_, object_path in memo.values():
        if not object_path.startswith('#/_references') and id_ in id_memo:
            if REF_MARKER not in serialized:
                set_in_object_from_path(dict_, object_path, {REF_MARKER: f'#/_references/{id_}'})


def serialize_sequence_with_pointers(seq, memo, path, id_method, id_memo):
    """ Serialize a sequence (list or tuple) using json-pointers. """
    serialized_sequence = []
    for ival, value in enumerate(seq):
        value_path = f'{path}/{ival}'
        serialized_value, memo = serialize_with_pointers(value, memo=memo, path=value_path, id_method=id_method,
                                                         id_memo=id_memo)
        serialized_sequence.append(serialized_value)
    return serialized_sequence, memo


def deserialize(serialized_element, sequence_annotation: str = 'List',
                global_dict=None, pointers_memo=None, path: str = '#'):
    """ Main function for deserialization, handle pointers. """
    if pointers_memo is not None:
        if path in pointers_memo:
            return pointers_memo[path]

    if isinstance(serialized_element, dict):
        return dict_to_object(serialized_element, global_dict=global_dict, pointers_memo=pointers_memo, path=path)
    if dcty.is_sequence(serialized_element):
        return deserialize_sequence(sequence=serialized_element, annotation=sequence_annotation,
                                    global_dict=global_dict, pointers_memo=pointers_memo, path=path)
    if isinstance(serialized_element, str):
        is_class_transformed = dcty.is_classname_transform(serialized_element)
        if is_class_transformed:
            return is_class_transformed
    return serialized_element


def deserialize_sequence(sequence, annotation=None, global_dict=None, pointers_memo=None, path='#'):
    """ Transform a sequence into an object. """
    origin, args = dcty.unfold_deep_annotation(typing_=annotation)
    deserialized_sequence = [deserialize(elt, args, global_dict=global_dict, pointers_memo=pointers_memo,
                                         path=f'{path}/{ie}') for ie, elt in enumerate(sequence)]

    if origin is tuple:
        # Keeping as a tuple
        return tuple(deserialized_sequence)
    return deserialized_sequence


def dict_to_object(dict_, class_=None, force_generic: bool = False, global_dict=None, pointers_memo=None, path='#'):
    """ Transform a dict into an object. """
    class_argspec = None

    if pointers_memo is None or global_dict is None:
        global_dict, pointers_memo = update_pointers_data(global_dict=global_dict, current_dict=dict_,
                                                          pointers_memo=pointers_memo)

    for marker in [REF_MARKER, OLD_REF_MARKER]:  # Retro-compatibility started on v0.13. When to remove?
        if marker in dict_:
            try:
                return pointers_memo[dict_[marker]]
            except KeyError as err:
                print('keys in memo:')
                for key in sorted(pointers_memo.keys()):
                    print(f'\t{key}')
                raise RuntimeError(f"Pointer {dict_[marker]} not in memo, at path {path}") from err

    if class_ is None and 'object_class' in dict_:
        class_ = get_python_class_from_class_name(dict_['object_class'])

    # Create init_dict
    if class_ is not None and hasattr(class_, 'dict_to_object'):
        different_methods = class_.dict_to_object.__func__ is not SerializableObject.dict_to_object.__func__
        if different_methods and not force_generic:
            try:
                obj = class_.dict_to_object(dict_, global_dict=global_dict, pointers_memo=pointers_memo, path=path)
            except TypeError:
                if issubclass(class_, SerializableObject):
                    warnings.warn(f'specific to_dict of class {class_.__name__} '
                                  'should implement use_pointers, memo and path arguments', Warning)
                obj = class_.dict_to_object(dict_)
            return obj

        if class_ in fullargsspec_cache:
            class_argspec = fullargsspec_cache[class_]
        else:
            class_argspec = inspect.getfullargspec(class_)
            fullargsspec_cache[class_] = class_argspec

        init_dict = {k: v for k, v in dict_.items() if k in class_argspec.args + class_argspec.kwonlyargs}
        # TOCHECK Class method to generate init_dict ??
    else:
        init_dict = dict_

    subobjects = {}
    for key, value in init_dict.items():
        if class_argspec is not None and key in class_argspec.annotations:
            annotation = class_argspec.annotations[key]
        else:
            annotation = None

        key_path = f'{path}/{key}'

        if key_path in pointers_memo:
            subobjects[key] = pointers_memo[key_path]
        else:
            subobjects[key] = deserialize(value, annotation, global_dict=global_dict,
                                          pointers_memo=pointers_memo, path=key_path)  # , enforce_pointers=False)
    if class_ is not None:
        obj = class_(**subobjects)
    else:
        obj = subobjects
    return obj


def deserialize_with_type(type_, value):
    """ Useful when deserialization is not straightforward and more info is need (type_). """
    if type_ in dcty.TYPES_STRINGS.values():
        return literal_eval(type_)(value)
    if isinstance(type_, str):
        class_ = get_python_class_from_class_name(type_)
        if inspect.isclass(class_):
            return class_.dict_to_object(value)
        raise NotImplementedError(f'Cannot get class from name {type_}')

    if isinstance(type_, (list, tuple)):
        return [deserialize_with_type(t, v) for t, v in zip(type_, value)]
    if type_ is None:
        return value

    raise NotImplementedError(type_)


def deserialize_with_typing(type_, argument, global_dict=None, pointers_memo=None, path='#'):
    """ Deserialize an object with a typing info. """
    origin = get_origin(type_)
    args = get_args(type_)
    deserialized_arg = None
    if origin is Union:
        # Check for Union false Positive (Default value = None)
        if len(args) == 2 and type(None) in args:
            return deserialize_argument(type_=args[0], argument=argument)

        # Type union
        classes = list(args)
        instantiated = False
        while instantiated is False:
            # Find the last class in the hierarchy
            hierarchy_lengths = [len(cls.mro()) for cls in classes]
            max_length = max(hierarchy_lengths)
            children_class_index = hierarchy_lengths.index(max_length)
            children_class = classes[children_class_index]
            try:
                # Try to deserialize
                # Throws KeyError if we try to put wrong dict into
                # dict_to_object. This means we try to instantiate
                # a children class with a parent dict_to_object
                deserialized_arg = children_class.dict_to_object(argument, global_dict=global_dict,
                                                                 pointers_memo=pointers_memo, path=path)

                # If it succeeds we have the right
                # class and instantiated object
                instantiated = True
            except KeyError:
                # This is not the right class, we should go see the parent
                classes.remove(children_class)
    elif origin in [list, collections.abc.Iterator]:
        # Homogeneous sequences (lists)
        sequence_subtype = args[0]
        deserialized_arg = [deserialize_argument(sequence_subtype, arg) for arg in argument]
        if origin is collections.abc.Iterator:
            deserialized_arg = iter(deserialized_arg)

    elif origin is tuple:
        # Heterogeneous sequences (tuples)
        deserialized_arg = tuple(deserialize_argument(t, arg) for (t, arg) in zip(args, argument))
    elif origin is dict:
        # Dynamic dict
        deserialized_arg = argument
    elif origin is InstanceOf:
        classname = args[0]
        object_class = full_classname(object_=classname, compute_for='class')
        class_ = get_python_class_from_class_name(object_class)

        deserialized_arg = class_.dict_to_object(argument, global_dict=global_dict,
                                                 pointers_memo=pointers_memo, path=path)

    elif type_ == dcty.Type:
        deserialized_arg = dcty.is_classname_transform(argument)
    else:
        msg = "Deserialization of typing {} is not implemented"
        raise NotImplementedError(msg.format(type_))
    return deserialized_arg


def deserialize_argument(type_, argument, global_dict=None, pointers_memo=None, path='#'):
    """ Deserialize an argument of a function with the type. """
    if argument is None:
        return None

    if isinstance(argument, SerializableObject):
        return argument

    if is_typing(type_):
        return deserialize_with_typing(type_, argument)

    if type_ in [TextIO, BinaryIO] or isinstance(argument, (StringFile, BinaryFile)):
        return argument

    if type_ in TYPING_EQUIVALENCES:
        if isinstance(argument, type_):
            return argument
        if isinstance(argument, int) and type_ == float:
            # Explicit conversion in this case
            return float(argument)
        msg = f"Given built-in type and argument are incompatible: {type(argument)} and {type_} in {argument}"
        raise TypeError(msg)

    if type_ is Any:
        # Any type
        return argument
    if inspect.isclass(type_) and issubclass(type_, SerializableObject):
        # Custom classes
        return type_.dict_to_object(argument, global_dict=global_dict, pointers_memo=pointers_memo, path=path)

    if type_ == dcty.Type:
        return dcty.is_classname_transform(argument)

    if issubclass(type_, Measure):
        return argument
    raise TypeError(f"Deserialization of type {type_} is Not Implemented")


def find_references(value, path='#'):
    """
    Traverse recursively the value to find reference (pointers) in it.

    Calls recursively find_references_sequence and find_references_dict.
    """
    if isinstance(value, dict):
        return find_references_dict(value, path)
    if dcty.isinstance_base_types(value):
        return []
    if dcty.is_sequence(value):
        return find_references_sequence(value, path)
    if isinstance(value, (BinaryFile, StringFile)):
        return []
    if isinstance(value, SerializableObject):
        return []
    raise ValueError(value)


def find_references_sequence(seq, path):
    """ Find dc refs recursively in sequence. """
    if isinstance(seq, str):
        raise ValueError

    references = []
    for ie, element in enumerate(seq):
        path_value = f'{path}/{ie}'
        references.extend(find_references(element, path=path_value))
        # if value_nodes or value_edges:

    return references


def find_references_dict(dict_, path):
    """ Find dc refs recursively in dict. """
    if REF_MARKER in dict_:
        return [(path, dict_[REF_MARKER])]

    # Retro-compatibility. Remove at some point.
    if OLD_REF_MARKER in dict_:
        return [(path, dict_[OLD_REF_MARKER])]

    references = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = f'{path}/{key}'
            refs_value = find_references(value, path=path_value)
            references.extend(refs_value)
    return references


def pointer_graph(value):
    """
    Create a graph of sub-attributes of an object.

    Edges representing either:
     * the hierarchy of an sub attribute to an attribute
     * the pointer link between the 2 elements
    """
    nodes = set()
    path_edges = set()
    pointer_edges = set()

    # refs = []
    refs_by_path = {}
    references = find_references(value)

    for path, reference in references:
        # refs.append(reference)
        refs_by_path[path] = reference
        # Slitting path & reference to add missing nodes
        # For path
        path_segments = path.split('/')
        nodes.add(path_segments[0])
        previous_node = path_segments[0]
        for segment in path_segments[1:]:
            node = f'{previous_node}/{segment}'
            nodes.add(node)
            path_edges.add((previous_node, node))
            previous_node = node

        # For reference
        reference_segments = reference.split('/')
        nodes.add(reference_segments[0])
        previous_node = reference_segments[0]
        for segment in reference_segments[1:]:
            node = f'{previous_node}/{segment}'
            nodes.add(node)
            path_edges.add((previous_node, node))
            previous_node = node

        pointer_edges.add((path, reference))

    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(path_edges, path=True)
    graph.add_edges_from(pointer_edges, head_type='circle', color='#008000', path=False)

    refs_by_path = {r[0]: r[1] for r in references}
    extra_edges = []
    new_paths = {}
    for path, point_at in references:  # loop over all refs?
        current_node = point_at
        last_name_change = None

        while current_node != '#':
            if current_node in refs_by_path:

                extra_edges.append((path, refs_by_path[current_node]))
                last_name_change = (current_node, refs_by_path[current_node])

            # Going up
            current_node = '/'.join(current_node.split('/')[:-1])

        if last_name_change:
            # marking as new path from top level pointer
            segments = point_at.split('/')
            for i in range(len(segments) - 1):
                node = '/'.join(segments[:i + 2])
                new_path = node.replace(*last_name_change)
                if new_path != node:
                    new_paths[node] = new_path
                    extra_edges.append((node, new_paths[node]))

    graph.add_edges_from(extra_edges, path=False)

    nx.set_node_attributes(graph, new_paths, "equals")

    return graph


def update_pointers_data(global_dict, current_dict, pointers_memo):
    """
    Update pointers according to current dict.

    :returns: the global dict and a pointer memo
    """
    if global_dict is None or pointers_memo is None:
        global_dict = current_dict

        if pointers_memo is None:
            pointers_memo = {}

        pointers_memo.update(dereference_jsonpointers(current_dict))
    return global_dict, pointers_memo


def deserialization_order(dict_):
    """ Analyze a dict representing an object and give a deserialization order. """
    graph = pointer_graph(dict_)
    if '#' not in graph.nodes:
        return []

    cycles = list(nx.simple_cycles(graph))
    if cycles:
        for cycle in cycles:
            print(cycle)
        raise NotImplementedError('Cycles in jsonpointers not handled')

    order = list(explore_tree_from_leaves(graph))
    if '#' in order:
        order.remove('#')

    cleaned_order = []
    equals = nx.get_node_attributes(graph, 'equals')

    for node in order:
        if node in equals:
            if not equals[node] in cleaned_order:
                cleaned_order.append((node, equals[node]))
        else:
            cleaned_order.append((node, None))
    return cleaned_order


def dereference_jsonpointers(dict_):  # , global_dict):
    """
    Analyze given dict.

     Useful in order to:
    - find JSON pointers
    - deserialize them in the right order to respect pointers graph
    :returns: a dictionary with key the path of the item and the value is the python object
    """
    order = deserialization_order(dict_)
    pointers_memo = {}
    for node, reference in order:
        if reference is None:
            serialized_element = get_in_object_from_path(object_=dict_, path=node, evaluate_pointers=False)
            pointers_memo[node] = deserialize(serialized_element=serialized_element, global_dict=dict_,
                                              pointers_memo=pointers_memo, path=node)
        else:
            pointers_memo[node] = pointers_memo[reference]
    return pointers_memo


def pointer_graph_elements(value, path='#'):
    """ Compute graph. """
    if isinstance(value, dict):
        return pointer_graph_elements_dict(value, path)
    if dcty.isinstance_base_types(value):
        return [], []
    if dcty.is_sequence(value):
        return pointer_graph_elements_sequence(value, path)

    raise ValueError(value)


def pointer_graph_elements_sequence(seq, path='#'):
    """ Compute graph from sequence. """
    if isinstance(seq, str):
        raise ValueError

    edges = []
    nodes = []
    for ie, element in enumerate(seq):
        path_value = f'{path}/{ie}'
        value_nodes, value_edges = pointer_graph_elements(element, path=path_value)
        # if value_nodes or value_edges:

        nodes.append(path_value)
        nodes.extend(value_nodes)

        edges.append((path, path_value))
        edges.extend(value_edges)

    return nodes, edges


def pointer_graph_elements_dict(dict_, path='#'):
    """ Compute graph from dict. """
    if REF_MARKER in dict_:
        return [path, dict_[REF_MARKER]], [(path, dict_[REF_MARKER], True)]

    # Retro compatibility. To be remove in the future
    if OLD_REF_MARKER in dict_:
        return [path, dict_[OLD_REF_MARKER]], [(path, dict_[OLD_REF_MARKER], True)]

    edges = []
    nodes = []
    for key, value in dict_.items():
        if not dcty.isinstance_base_types(value):
            path_value = f'{path}/{key}'
            value_nodes, value_edges = pointer_graph_elements(value, path=path_value)
            # if value_nodes or value_edges:
            nodes.append(path_value)
            nodes.extend(value_nodes)

            edges.append((path, path_value, False))
            edges.extend(value_edges)

    return nodes, edges


def pointers_analysis(obj):
    """
    Analyze object to output stats on pointer use in it.

    Maybe useless now that we use uuids.
    :returns: a tuple of 2 dictionaries: one giving the number of pointer use by class
    """
    if isinstance(obj, dict):
        dict_ = obj
    else:
        dict_ = obj.to_dict()

    class_number = {}
    composed_by = {}
    class_from_path = {}
    graph = pointer_graph(dict_)
    for path1, path2 in graph.edges():
        if path1 != '#':
            if path2 in class_from_path:
                val2_class = class_from_path[path2]
            else:
                try:
                    val2 = get_in_object_from_path(obj, path2)
                    val2_class = full_classname(val2)
                    class_from_path[path2] = val2_class
                except AttributeError:
                    val2_class = None

            if val2_class:
                class_number[val2_class] = class_number.get(val2_class, 0) + 1

            if path1 in class_from_path:
                val1_class = class_from_path[path1]
            else:
                try:
                    val1 = get_in_object_from_path(obj, path1)
                    val1_class = full_classname(val1)
                    class_from_path[path1] = val1_class
                except AttributeError:
                    val1_class = None

            if val1_class != val2_class:
                if val2_class not in composed_by:
                    composed_by[val2_class] = {}

                if val1_class and val1_class not in composed_by[val2_class]:
                    composed_by[val2_class][val1_class] = 1
                else:
                    composed_by[val2_class][val1_class] += 1

    return class_number, composed_by


def is_serializable(obj):
    """ Check whether given object is serializable. """
    if dcty.is_jsonable(obj):
        return True
    if isinstance(obj, SerializableObject):
        dict_ = obj.to_dict(use_pointers=False)
        return dcty.is_jsonable(dict_)
    if isinstance(obj, dict):
        for key, value in obj.items():
            if not is_serializable(key) or not is_serializable(value):
                return False
        return True
    if dcty.is_sequence(obj):
        for element in obj:
            if not is_serializable(element):
                return False
        return True
    return False




import unittest

scripts = [
    # Framework
    "displays.py",
    "generation.py",
    "models_test.py",
    "files.py",
    "clustering.py",
    "dataset.py",
    "dataset_pareto.py",
    "filters.py",
    "optimization.py",
    "graph.py",
    "sampling.py",
    "markdowns.py",
    "checks.py",
    "docx_writer.py",
    "markdown_to_docx.py",

    # Workflows
    "workflow/power_simulation.py",
    "workflow/workflow_diverge_converge.py",
    "workflow/workflow_clustering.py",
    "workflow/workflow_filtering.py",
    "workflow/workflow_pareto.py",
    "workflow/workflow_building.py",
    "workflow/workflow_sampling.py",
    "workflow/workflow_inputs.py",
    "workflow/file_inputs.py",

    # Breakdown
    "breakdown.py",

    # Utils
    "utils/algebra.py",
    "utils/interpolation.py",
    "utils/serialization.py",
    "utils/helpers.py",
    "utils/diff.py",

    # Data structures
    "data-structures/1.py"
]

for script_name in scripts:
    print(f"\n## Executing script '{script_name}'.")
    with open(script_name, "r") as script_file:
        exec(script_file.read())
        print(f"Script '{script_name}' successful.")

# This needs to be executed once all "assert-tests" have been run + once all unittests are defined
if __name__ == "__main__":
    unittest.main(verbosity=3)



"""
Test for pareto in workflows
"""
import importlib_resources
import json
from dessia_common.files import StringFile
from dessia_common.typings import ClassMethodType, MethodType
from dessia_common.tests import Car
from dessia_common.workflow.blocks import ClassMethod, InstantiateModel, ModelMethod
from dessia_common.datatools.dataset import Dataset
from dessia_common.workflow.core import Workflow, Pipe

# Import data
ref = importlib_resources.files("dessia_common").joinpath("models/data/cars.csv")
with ref.open('rb') as csv_cars:
    stream = StringFile.from_stream(csv_cars)

    # Workflow
    block_0 = ClassMethod(method_type=ClassMethodType(Car, 'from_csv'), name='CSV Cars')
    block_1 = InstantiateModel(model_class=Dataset, name='Dataset')
    block_2 = ModelMethod(method_type=MethodType(Dataset, 'pareto_points'), name='Pareto Points')
    blocks = [block_0, block_1, block_2]

    pipe_0 = Pipe(block_0.outputs[0], block_1.inputs[0])
    pipe_2 = Pipe(block_1.outputs[0], block_2.inputs[0])
    pipes = [pipe_0, pipe_2]

    workflow = Workflow(blocks, pipes, output=block_2.outputs[0], name='Pick Pareto members')

    # Workflow run
    workflow_run = workflow.run({
        workflow.index(block_0.inputs[0]): stream,
        workflow.index(block_2.inputs[1]): ['weight', 'mpg']})

# Workflow tests
json_dict = json.dumps(workflow.to_dict())
decoded_json = json.loads(json_dict)
deserialized_object = workflow.dict_to_object(decoded_json)
workflow._check_platform()

workflow._check_platform()
wfrun_plot_data = workflow_run.output_value.plot_data()
# assert(json.dumps(wfrun_plot_data[0].to_dict())[50:100] == 's": [{"mpg": 26.0, "cylinders": 4.0, "displacement')
# assert(json.dumps(wfrun_plot_data[1].to_dict())[10400:10450] == ': ["mpg", "cylinders", "displacement", "horsepower')
# assert(json.dumps(wfrun_plot_data[2].to_dict())[50:100] == 'te_names": ["Index of reduced basis vector", "Sing')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

"""

from dessia_common.generation import RegularDecisionTreeGenerator, DecisionTreeGenerator


class Model:
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return 'Model of items: {}'.format(self.items)

    def is_valid(self):
        if len(self.items) >= 2:
            if self.items[1] == 0:
                return False
        return True


class RegularCustomModelGenerator(RegularDecisionTreeGenerator):
    def __init__(self, name: str = ''):
        RegularDecisionTreeGenerator.__init__(self, number_possibilities=[2, 2, 2], name=name)

    def model_from_vector(self, vector):
        return Model(vector)

    def is_model_valid(self, model):
        return model.is_valid()


class CustomModelGenerator(DecisionTreeGenerator):
    def __init__(self, name: str = ''):
        DecisionTreeGenerator.__init__(self, name=name)

    def model_from_vector(self, vector):
        return Model(vector)

    def is_model_valid(self, model):
        return model.is_valid()

    def number_possibilities_from_model(self, model):
        if len(model.items) < 3:
            return 2
        return 0


generator = CustomModelGenerator()

models = []
for model in generator.generate(verbose=True):
    print(model)
    models.append(model)

assert len(models) == 4

regular_generator = RegularCustomModelGenerator()

regular_models = []
for model in regular_generator.generate(verbose=True):
    print(model)
    regular_models.append(model)

assert len(regular_models) == 4



"""
    test markdown to docx
"""

from dessia_common.models import all_cars_no_feat
from dessia_common.datatools import HeterogeneousList
from dessia_common.document_generator import DocxWriter

dataset = HeterogeneousList(all_cars_no_feat)
docx_writer = DocxWriter.from_markdown(markdown_text=dataset.to_markdown())
docx_writer.save_file()

print("script 'markdown_to_docx.py' has passed")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Compilation test
"""

import sys


# =============================================================================
#  Tests
# =============================================================================

if sys.argv[1] == 'test_valid_license':
    license_should_be_valid = True
    print('Testing for valid license')
elif sys.argv[1] == 'test_unvalid_license':
    license_should_be_valid = False
    print('Testing for unvalid license')
else:
    raise ValueError('Invalid option: use either test_valid_license ot test_unvalid_license')


test_passed = license_should_be_valid
try:
    from dessia_common.utils.helpers import prettyname
    prettyname('aa')
    print('No license error triggered')
except RuntimeError:
    print('License error triggered')
    if license_should_be_valid:
        print('License was expected to be valid, test failed')
        test_passed = False
    else:
        print('License was expected to be unvalid, test passed')
        test_passed = True

assert test_passed



"""
Tools for objects handling.

As a rule of thumb, functions should be placed here if:
- They can be widely used in dessia_common
- They don't have any requirements or imports

That way, we can avoid cyclic imports.
"""

import sys
from importlib import import_module
from ast import literal_eval
from typing import Type

_PYTHON_CLASS_CACHE = {}


def concatenate(values):
    """ Concatenate values of class class_ into a class_ containing all concatenated values. """
    types_set = set(type(value) for value in values)
    concatenated_values = None
    if len(types_set) != 1:
        raise TypeError("Block Concatenate only defined for operands of the same type.")

    first_value = values[0]
    if hasattr(first_value, 'extend'):
        concatenated_values = first_value.__class__()
        for value in values:
            concatenated_values.extend(value)

    if concatenated_values is not None:
        return concatenated_values

    raise ValueError("Block Concatenate only defined for classes with 'extend' method")


def prettyname(name: str) -> str:
    """ Create a pretty name from a string. """
    pretty_name = ''
    if name:
        strings = name.split('_')
        for i, string in enumerate(strings):
            if len(string) > 1:
                pretty_name += string[0].upper() + string[1:]
            else:
                pretty_name += string
            if i < len(strings) - 1:
                pretty_name += ' '
    return pretty_name


def full_classname(object_, compute_for: str = 'instance') -> str:
    """ Get full class name of object_ (module + class name). """
    if compute_for == 'instance':
        return f"{object_.__class__.__module__}.{object_.__class__.__name__}"
    if compute_for == 'class':
        return f"{object_.__module__}.{object_.__name__}"
    raise NotImplementedError(f"Cannot compute '{compute_for}' full classname for object '{object_}'")


def get_python_class_from_class_name(full_class_name: str) -> Type:
    """ Get python class object corresponding to given class name. """
    cached_value = _PYTHON_CLASS_CACHE.get(full_class_name, None)
    sys.setrecursionlimit(3000)  # TODO : this is just quick fix, we haven't found the real reason behind it.
    if cached_value is not None:
        return cached_value

    if "." not in full_class_name:
        return literal_eval(full_class_name)
    module_name, class_name = full_class_name.rsplit(".", 1)

    module = import_module(module_name)
    class_ = getattr(module, class_name)
    # Storing in cache
    _PYTHON_CLASS_CACHE[full_class_name] = class_
    return class_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Module to handle serialization for engineering objects. """

import time
import warnings
import operator
import math
import random
import itertools
import zipfile

from copy import deepcopy, copy
import inspect
import json

from typing import List, Tuple
import traceback as tb

from importlib import import_module

from dessia_common.utils.diff import data_eq, diff, choose_hash
from dessia_common.utils.types import is_bson_valid
from dessia_common.utils.copy import deepcopy_value
import dessia_common.schemas.core as dcs
from dessia_common.serialization import SerializableObject, deserialize_argument, serialize
from dessia_common.exports import XLSXWriter, MarkdownWriter, ExportFormat
from dessia_common.typings import JsonSerializable
from dessia_common import templates
import dessia_common.checks as dcc
from dessia_common.displays import DisplayObject, DisplaySetting
from dessia_common.breakdown import attrmethod_getter, get_in_object_from_path
import dessia_common.utils.helpers as dch
import dessia_common.files as dcf
from dessia_common.document_generator import DocxWriter
from dessia_common.decorators import get_decorated_methods, DISPLAY_DECORATORS


def __getattr__(name):
    if name == "_FORBIDDEN_ARGNAMES":
        warnings.warn("Attribute '_FORBIDDEN_ARGNAMES' is deprecated. Use schemas.RESERVED_ARGNAMES instead",
                      DeprecationWarning)
        return dcs.RESERVED_ARGNAMES
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


_fullargsspec_cache = {}


class DessiaObject(SerializableObject):
    """
    Base class for Dessia's platform compatible objects.

    Gathers generic methods and attributes

    :cvar bool _standalone_in_db: Indicates whether class objects should be independent in database or not.
        If False, object will only exist inside its parent.

    :cvar bool _eq_is_data_eq:
        Indicates which type of equality check is used: strict equality or equality based on data.
        If False, Python's object __eq__ method is used (strict), else, user custom data_eq is used (data)

    :cvar List[str] _non_serializable_attributes:
        [Advanced] List of instance attributes that should not be part of serialization with to_dict method.
        These will not be displayed in platform object tree, for instance.

    :cvar List[str] _non_data_eq_attributes:
        [Advanced] List of instance attributes that should not be part of equality check with data__eq__ method
        (if _eq_is_data_eq is True).

    :cvar List[str] _non_data_hash_attributes:
        [Advanced] List of instance attributes that should not be part of hash computation with data__hash__ method
        (if _eq_is_data_eq is True).

    :cvar List[str] _allowed_methods: List of all methods that can be ran from platform.

    :param name: Name of object.
    """

    _non_editable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _eq_is_data_eq = True
    _vector_features = None

    _init_variables = None
    _allowed_methods = []

    def __init__(self, name: str = '', **kwargs):
        self.name = name
        if kwargs:
            warnings.warn(('Providing attributes to DessiaObject __init__ to be stored in self is deprecated\n'
                           + 'Please store your attributes by yourself in your init'),
                          DeprecationWarning)

        # The code below has shown to be inefficient and should be remove in future version (0.16?)
        for property_name, property_value in kwargs.items():
            setattr(self, property_name, property_value)

    def base_dict(self):
        """ Base dict of the object, with just its name. """
        dict_ = SerializableObject.base_dict(self)
        dict_['name'] = self.name
        return dict_

    def __hash__(self):
        """ Compute a int from object. """
        if self._eq_is_data_eq:
            return self._data_hash()
        return object.__hash__(self)

    def __eq__(self, other_object):
        """
        Generic equality of two objects.

        Behavior can be controlled by class attribute _eq_is_data_eq to tell if we must use python equality (based on
        memory addresses) (_eq_is_data_eq = False) or a data equality (True).
        """
        if hash(self) != hash(other_object):
            return False
        if self._eq_is_data_eq:
            if self.__class__.__name__ != other_object.__class__.__name__:
                return False
            return self._data_eq(other_object)
        return object.__eq__(self, other_object)

    def _data_eq_dict(self):
        """ Returns a dict of what to look at for data eq. Keys in non data eq attributes are removed. """
        return {k: v for k, v in self._serializable_dict().items()
                if k not in self._non_data_eq_attributes + ['package_version', 'name']}

    def _data_eq(self, other_object) -> bool:
        """ Returns if the object is equal to the other object in the sense of data contained in the objects. """
        return data_eq(self, other_object)

    def _data_hash(self):
        """ Generic computation of hash based on data. """
        forbidden_keys = (self._non_data_eq_attributes + self._non_data_hash_attributes + ['package_version', 'name'])
        hash_ = sum(choose_hash(v) for k, v in self._serializable_dict().items() if k not in forbidden_keys)
        return int(hash_ % 1e5)

    def _data_diff(self, other_object):
        """
        Make a diff between two objects.

        returns: different values, missing keys in other object
        """
        return diff(self, other_object)

    def _get_from_path(self, path: str):
        """ Get object's deep attribute from given path. """
        return get_in_object_from_path(self, path)

    @classmethod
    def raw_schema(cls):
        """ Schema of class: transfer python data structure to web standard. """
        if hasattr(cls, '_jsonschema'):
            warnings.warn("Jsonschema is fully deprecated and you may want to use the new generic schema feature."
                          "Please consider so", DeprecationWarning)
            return cls._jsonschema
        return dcs.ClassSchema(cls)

    @classmethod
    def schema(cls):
        """ Serialized Schema as a dict. """
        return cls.raw_schema().to_dict()

    @classmethod
    def jsonschema(cls):
        """ Jsonschema of class: transfer python data structure to web standard. """
        warnings.warn("'jsonschema' method is deprecated. Use schema instead", DeprecationWarning)
        return cls.schema()

    @property
    def raw_method_schemas(self):
        """ Generate dynamic schemas for methods of class. """
        # TODO This should be a classmethod, but is a property to avoid collision with workflow.
        cls = self.__class__
        valid_method_names = [m for m in dir(cls) if not m.startswith('_') and m in cls._allowed_methods]
        schemas = {}
        for method_name in valid_method_names:
            method = getattr(cls, method_name)
            schemas[method_name] = dcs.MethodSchema(method)
        return schemas

    @property
    def method_schemas(self):
        """ Generate dynamic schemas for methods of class. """
        return {method_name: schema.to_dict() for method_name, schema in self.raw_method_schemas.items()}

    @property
    def _method_jsonschemas(self):
        """ Generates dynamic schemas for methods of class. """
        warnings.warn("method_jsonschema method is deprecated. Use method_schema instead", DeprecationWarning)
        return self.method_schemas

    def method_dict(self, method_name: str, method_jsonschema=None):
        """ Return a default dict for the given method name. """
        if method_jsonschema is not None:
            warnings.warn("method_jsonschema argument is deprecated and its use will be removed in a future version."
                          " Please remove it from your function call. Method name is sufficient to get schema",
                          DeprecationWarning)
        schema = self.raw_method_schemas[method_name]
        return schema.default_dict()

    def dict_to_arguments(self, dict_, method):
        """ Transform serialized argument of a method to python objects ready to use in method evaluation. """
        method_full_name = f"{self.full_classname}.{method}"
        if method_full_name in _fullargsspec_cache:
            args_specs = _fullargsspec_cache[method_full_name]
        else:
            method_object = getattr(self, method)
            args_specs = inspect.getfullargspec(method_object)
            _fullargsspec_cache[method_full_name] = args_specs

        allowed_args = args_specs.args[1:]

        arguments = {}
        for i, arg in enumerate(allowed_args):
            if str(i) in dict_:
                arg_specs = args_specs.annotations[arg]
                value = dict_[str(i)]
                try:
                    deserialized_value = deserialize_argument(arg_specs, value)
                except TypeError as err:
                    msg = "Error in deserialisation of value: "
                    msg += f"{value} of expected type {arg_specs}"
                    raise TypeError(msg) from err
                arguments[arg] = deserialized_value
        return arguments

    def save_to_file(self, filepath: str, indent: int = 2):
        """
        Save object to a JSON file.

        :param filepath: either a string representing the filepath or a stream
        """
        if not filepath.endswith('.json'):
            filepath += '.json'
            print(f'Changing name to {filepath}')
        with open(filepath, 'w', encoding='utf-8') as file:
            self.save_to_stream(file, indent=indent)

    def save_to_stream(self, stream, indent: int = 2):
        """ Write object to a stream. Default is to_dict, as a text like stream. """
        try:
            dict_ = self.to_dict(use_pointers=True)
        except TypeError:
            dict_ = self.to_dict()

        json.dump(dict_, stream, indent=indent)
        stream.write('\n')

    @classmethod
    def load_from_stream(cls, stream: dcf.JsonFile):
        """
        Generate object from stream using utf-8 encoding.

        Should be consistent with save_to_stream method.
        """
        warnings.warn("This method is deprecated and will be removed in a future version."
                      " Please use the `from_json_stream` method instead.", DeprecationWarning)
        return cls.from_json_stream(stream=stream)

    @classmethod
    def load_from_file(cls, filepath: str):
        """
        Load object from a JSON file.

        :param filepath: either a string representing the filepath or a stream
        """
        warnings.warn("This method is deprecated and will be removed in a future version."
                      " Please use the `from_json` method instead.", DeprecationWarning)
        return cls.from_json(filepath=filepath)

    @classmethod
    def from_json_stream(cls, stream: dcf.JsonFile):
        """
        Generate object from stream using utf-8 encoding.

        Should be consistent with save_to_stream method.
        """
        dict_ = json.loads(stream.read().decode('utf-8'))
        return cls.dict_to_object(dict_)

    @classmethod
    def from_json(cls, filepath: str):
        """
        Load object from a JSON file.

        :param filepath: either a string representing the filepath or a stream
        """
        with open(filepath, 'r', encoding='utf-8') as file:
            dict_ = json.load(file)

        return cls.dict_to_object(dict_)

    def check_list(self, level: str = 'error', check_platform: bool = True) -> dcc.CheckList:
        """ Return a list of potential info, warning and issues on the instance, that might be user custom. """
        check_list = dcc.CheckList([])

        if check_platform:
            check_list += self.check_platform()
        return check_list

    def is_valid(self, level: str = 'error') -> bool:
        """
        Return whether the object of valid 'above' given level.

        Default is error, but warnings can be forbidden.
        """
        return not self.check_list().checks_above_level(level=level)

    def copy(self, deep: bool = True, memo=None):
        """
        Copy object.

        :param deep: If False, perform a shallow copy. If True, perform a deep copy.
        :param memo: A dict that keep track of references.
        """
        if deep:
            return deepcopy(self, memo=memo)
        return copy(self)

    def __copy__(self):
        """ Generic copy use init of objects. """
        class_name = self.full_classname
        if class_name in _fullargsspec_cache:
            class_argspec = _fullargsspec_cache[class_name]
        else:
            class_argspec = inspect.getfullargspec(self.__class__)
            _fullargsspec_cache[class_name] = class_argspec

        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                value = self.__dict__[arg]
                if hasattr(value, '__copy__'):
                    dict_[arg] = value.__copy__()
                else:
                    dict_[arg] = value
        return self.__class__(**dict_)

    def __deepcopy__(self, memo=None):
        """ Generic deep copy use init functions of objects. """
        class_name = self.full_classname
        if class_name in _fullargsspec_cache:
            class_argspec = _fullargsspec_cache[class_name]
        else:
            class_argspec = inspect.getfullargspec(self.__class__)
            _fullargsspec_cache[class_name] = class_argspec

        if memo is None:
            memo = {}
        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                dict_[arg] = deepcopy_value(getattr(self, arg), memo=memo)
        return self.__class__(**dict_)

    def plot_data(self, reference_path: str = "#", **kwargs):
        """
        Base plot_data method. Overwrite this to display 2D or graphs on platform.

        Should return a list of plot_data's objects.
        """
        warnings.warn("Function 'plot_data' is deprecated and should not be used to compute graphs and 2D displays,"
                      "anymore. Please use Display Decorators, instead", DeprecationWarning)
        return []

    def plot(self, reference_path: str = "#", **kwargs):
        """ Generic plot getting plot_data function to plot. """
        if hasattr(self, 'plot_data'):
            import plot_data
            for data in self.plot_data(reference_path, **kwargs):
                plot_data.plot_canvas(plot_data_object=data,
                                      canvas_id='canvas',
                                      width=1400, height=900,
                                      debug_mode=False)
        else:
            msg = f"Class '{self.__class__.__name__}' does not implement a plot_data method to define what to plot"
            raise NotImplementedError(msg)

    def mpl_plot(self, **kwargs):
        """ Plot with matplotlib using plot_data function. """
        axs = []
        if hasattr(self, 'plot_data'):
            try:
                plot_datas = self.plot_data(**kwargs)
            except TypeError as error:
                raise TypeError(f'{self.__class__.__name__}.{error}') from error
            for data in plot_datas:
                if hasattr(data, 'mpl_plot'):
                    ax = data.mpl_plot()
                    axs.append(ax)
        else:
            msg = f"Class '{self.__class__.__name__}' does not implement a plot_data method to define what to plot"
            raise NotImplementedError(msg)
        return axs

    @classmethod
    def display_settings(cls, **kwargs) -> List[DisplaySetting]:
        """ Return a list of objects describing how to call object displays. """
        settings = [DisplaySetting(selector="markdown", type_="markdown", method="to_markdown", load_by_default=True)]
        settings.extend(cls._display_settings_from_decorators())
        return settings

    @classmethod
    def _display_settings_from_decorator_name(cls, decorator_name: str):
        methods = get_decorated_methods(class_=cls, decorator_name=decorator_name)
        settings = []
        for method in methods:
            name = method.__name__
            type_ = getattr(method, "type_", False)
            serialize_data = getattr(method, "serialize_data", False)
            load_by_default = getattr(method, "load_by_default", False)
            selector = getattr(method, "selector", None)
            if selector is None:
                selector = name
            settings.append(DisplaySetting(selector=selector, type_=type_, method=name,
                                           serialize_data=serialize_data, load_by_default=load_by_default))
        return settings
    
    @classmethod
    def _display_settings_from_decorators(cls) -> List[DisplaySetting]:
        """ Return a list, computed from decorated functions, of objects describing how to call displays. """
        return [s for d in DISPLAY_DECORATORS for s in cls._display_settings_from_decorator_name(d)]

    def _display_from_selector(self, selector: str) -> DisplayObject:
        """ Generate the display from the selector. """
        display_setting = self._display_settings_from_selector(selector)
        track = ""
        try:
            data = attrmethod_getter(self, display_setting.method)(**display_setting.arguments)
        except:
            data = None
            track = tb.format_exc()

        if display_setting.serialize_data:
            data = serialize(data)
        reference_path = display_setting.reference_path  # Trying this
        return DisplayObject(type_=display_setting.type, data=data, reference_path=reference_path, traceback=track)

    def _display_settings_from_selector(self, selector: str):
        """ Get display settings from given selector. """
        for display_setting in self.display_settings():
            if display_setting.selector == selector:
                return display_setting
        raise ValueError(f"No such selector '{selector}' in display of class '{self.__class__.__name__}'")

    def _displays(self) -> List[JsonSerializable]:
        """ Generate displays of the object to be plot in the DessiA Platform. """
        displays = []
        for display_setting in self.display_settings():
            display_ = self._display_from_selector(display_setting.selector)
            displays.append(display_.to_dict())
        return displays

    def to_markdown(self, **kwargs) -> str:
        """ Render a markdown of the object output type: string. """
        writer = MarkdownWriter(print_limit=25, table_limit=None)
        template = templates.dessia_object_markdown_template
        return template.substitute(name=self.name, class_=self.__class__.__name__, table=writer.object_table(self))

    def performance_analysis(self):
        """ Print time of rendering some commons operations (serialization, hash, displays). """
        print(f'### Performance analysis of object {self} ###')
        data_hash_time = time.time()
        self._data_hash()
        data_hash_time = time.time() - data_hash_time
        print(f'\t- data hash time: {round(data_hash_time, 3)} seconds')

        todict_time = time.time()
        dict_ = self.to_dict()
        todict_time = time.time() - todict_time
        print(f'\t- to_dict time: {round(todict_time, 3)} seconds')

        dto_time = time.time()
        self.dict_to_object(dict_)
        dto_time = time.time() - dto_time
        print(f'\t- dict_to_object time: {round(dto_time, 3)} seconds')

        for display_setting in self.display_settings():
            display_time = time.time()
            self._display_from_selector(display_setting.selector)
            display_time = time.time() - display_time
            print(f'\t- generation of display {display_setting.selector} in: {round(display_time, 6)} seconds')
        print('\n')

    def _check_platform(self, level='error'):
        return self.check_platform().raise_if_above_level(level=level)

    def check_platform(self):
        """ Reproduce life-cycle on platform (serialization, display). Raise an error if something is wrong. """
        serializable_results = dcc.check_serialization_process(object_=self, use_pointers=True)
        dict_ = serializable_results.pop("dict_")

        copy_results = dcc.check_copy(self)

        # Not refactoring this due to dependencies and cyclic imports
        print("Checking BSON validity...")
        start = time.time()
        valid, hint = is_bson_valid(stringify_dict_keys(dict_))
        if not valid:
            print("Failed.\n")
            check = dcc.FailedCheck(f"Object is not BSON valid {hint}")
        else:
            print("Object is BSON valid.")
            check = dcc.PassedCheck("Object is BSON valid")
        duration = time.time() - start
        print(f"Checked BSON validity in {duration}s.\n")
        bson_results = {"check": check, "duration": duration}

        display_results = dcc.check_displays(self)
        schemas_results = dcc.check_schemas(self)

        results = [serializable_results, copy_results, bson_results, display_results, schemas_results]
        duration = sum(r["duration"] for r in results)
        print(f"\nCompleted Platform Check in {duration}s.\n")
        return dcc.CheckList([r["check"] for r in results])

    def to_xlsx(self, filepath: str):
        """ Export the object to an XLSX file given by the filepath. """
        if not filepath.endswith('.xlsx'):
            filepath += '.xlsx'
            print(f'Changing name to {filepath}')

        with open(filepath, 'wb') as file:
            self.to_xlsx_stream(file)

    def to_xlsx_stream(self, stream):
        """ Export the object to an XLSX to a given stream. """
        writer = XLSXWriter(self)
        writer.save_to_stream(stream)

    def to_markdown_docx(self):
        """ Generates a word document from the object attributes, using a markdown-like syntax. """
        docx_writer = DocxWriter.from_markdown(markdown_text=self.to_markdown())
        return docx_writer

    def to_docx_stream(self, stream: dcf.BinaryFile):
        """ Saves the document to a binary stream. """
        document = self.to_markdown_docx()
        document.save_to_stream(stream=stream)

    def to_docx(self, filepath: str):
        """ Saves the document to a file. """
        if not filepath.endswith('.docx'):
            filepath += '.docx'
            print(f'Changing name to {filepath}')
        self.to_docx_stream(filepath)

    def zip_settings(self):
        """
        Returns a list of streams containing different representations of the object.

        Excel file stream generated by calling 'to_xlsx_stream' method.
        JSON file stream generated by calling 'save_to_stream' method.
        """
        excel_stream = dcf.BinaryFile("excel_export.xlsx")
        self.to_xlsx_stream(excel_stream)

        json_stream = dcf.JsonFile("json_export.json")
        self.save_to_stream(json_stream)

        return [excel_stream, json_stream]

    def to_zip_stream(self, archive: dcf.BinaryFile) -> List[dcf.BinaryFile]:
        """ Generates a zipped archive of the streams returned by 'zip_settings' method. """
        streams = self.zip_settings()
        archive_name = 'export_zip'
        archive.filename = archive_name
        with zipfile.ZipFile(archive, 'w', compression=zipfile.ZIP_DEFLATED) as zip_archive:
            for value in streams:
                archive = dcf.generate_archive(zip_archive=zip_archive, value=value)
        return [archive]

    def to_zip(self, filepath: str):
        """ Saves the zipped archive generated by 'to_zip_stream' method to a file. """
        if not filepath.endswith('.zip'):
            filepath += '.zip'
            print(f'Changing name to {filepath}')
        archive = dcf.BinaryFile()
        self.to_zip_stream(archive)
        with open(filepath, 'wb') as file:
            file.write(archive.getbuffer())

    def _export_formats(self) -> List[ExportFormat]:
        """ Return a list of objects describing how to call generic exports (.json, .xlsx). """
        formats = [ExportFormat(selector="json", extension="json", method_name="save_to_stream", text=True),
                   ExportFormat(selector="xlsx", extension="xlsx", method_name="to_xlsx_stream", text=False),
                   ExportFormat(selector="zip", extension="zip", method_name="to_zip_stream", text=False),
                   ExportFormat(selector="docx", extension="docx", method_name="to_docx_stream", text=False)]
        return formats

    def save_export_to_file(self, selector: str, filepath: str):
        """ Generic export from selector to given filepath. Return real location filepath. """
        for export_format in self._export_formats():
            if export_format.selector == selector:
                if not filepath.endswith(f".{export_format.extension}"):
                    filepath += f".{export_format.extension}"
                    print(f'Renaming filepath to {filepath}')
                if export_format.text:
                    with open(filepath, mode="w", encoding="utf-8") as stream:
                        getattr(self, export_format.method_name)(stream, **export_format.args)
                else:
                    with open(filepath, mode="wb") as stream:
                        getattr(self, export_format.method_name)(stream, **export_format.args)
                return filepath
        raise ValueError(f'Export selector not found: {selector}')

    def to_vector(self):
        """ Compute vector from object. """
        vectored_objects = []
        for feature in self.vector_features():
            vectored_objects.append(get_in_object_from_path(self, feature.lower()))
        return vectored_objects

    @classmethod
    def vector_features(cls):
        """ Get a list of vector features, or generate a default one. """
        if cls._vector_features is None:
            return list(set(get_attribute_names(cls)).difference(get_attribute_names(DessiaObject)))
        return cls._vector_features


class PhysicalObject(DessiaObject):
    """ Represent an object with CAD capabilities. """

    @classmethod
    def display_settings(cls, **kwargs):
        """ Returns a list of DisplaySettings objects describing how to call sub-displays. """
        display_settings = super().display_settings()
        display_settings.append(DisplaySetting(selector='cad', type_='babylon_data',
                                               method='volmdlr_volume_model().babylon_data', serialize_data=True))
        return display_settings

    def volmdlr_primitives(self, **kwargs):
        """ Return a list of volmdlr primitives to build up volume model. """
        return []

    def volmdlr_volume_model(self, **kwargs):
        """ Gives the volmdlr VolumeModel. """
        import volmdlr as vm  # !!! Avoid circular imports, is this OK ?
        return vm.core.VolumeModel(self.volmdlr_primitives(**kwargs))

    def to_step(self, filepath: str):
        """
        Exports the CAD of the object to step. Works if the class define a custom volmdlr model.

        :param filepath: a str representing a filepath
        """
        return self.volmdlr_volume_model().to_step(filepath=filepath)

    def to_step_stream(self, stream):
        """
        Export object CAD to given stream in STEP format.

        Works if the class define a custom volmdlr model.
        """
        return self.volmdlr_volume_model().to_step_stream(stream=stream)

    def to_html_stream(self, stream: dcf.StringFile):
        """ Exports Object CAD to given stream as HTML. """
        model = self.volmdlr_volume_model()
        babylon_data = model.babylon_data()
        script = model.babylonjs_script(babylon_data)
        stream.write(script)

        return stream

    def to_stl_stream(self, stream):
        """ Export Object CAD to given stream as STL. """
        return self.volmdlr_volume_model().to_stl_stream(stream=stream)

    def to_stl(self, filepath):
        """
        Exports the CAD of the object to STL. Works if the class define a custom volmdlr model.

        :param filepath: a str representing a filepath
        """
        return self.volmdlr_volume_model().to_stl(filepath=filepath)

    def babylonjs(self, use_cdn=True, debug=False, **kwargs):
        """ Show the 3D volmdlr of an object by calling volmdlr_volume_model method and plot in in browser. """
        self.volmdlr_volume_model(**kwargs).babylonjs(use_cdn=use_cdn, debug=debug)

    def save_babylonjs_to_file(self, filename: str = None, use_cdn: bool = True, debug: bool = False, **kwargs):
        """
        Saves the 3D volmdlr of an object in a file.

        :param filename: The file's name. Default value is None
        :type filename: str, optional
        :param use_cdn: Activates the use of a content delivery network. Default value is True
        :type use_cdn: bool, optional
        :param debug: Activates the debug mode. Default value is False
        :type debug: bool, optional
        """
        self.volmdlr_volume_model(**kwargs).save_babylonjs_to_file(filename=filename, use_cdn=use_cdn, debug=debug)

    def zip_settings(self):
        """
        Returns a list of streams containing different representations of the object.

        Excel file stream generated by calling 'to_xlsx_stream' method.
        JSON file stream generated by calling 'save_to_stream' method.
        STEP file stream generated by calling 'to_step_stream' method.
        HTML file stream generated by calling 'to_html_stream' method.
        STL file stream generated by calling 'to_stl_stream' method.
        """
        streams = DessiaObject.zip_settings(self)
        step_stream = dcf.StringFile("step_export.stp")
        self.to_step_stream(step_stream)
        html_stream = self.to_html_stream(dcf.StringFile(filename="html_export.html"))
        stl_stream = dcf.BinaryFile("stl_export.stl")
        self.to_stl_stream(stl_stream)

        streams.extend([step_stream, html_stream, stl_stream])
        return streams

    def _export_formats(self) -> List[ExportFormat]:
        """ Return a list of objects describing how to call 3D exports. """
        formats = DessiaObject._export_formats(self)
        formats3d = [ExportFormat(selector="step", extension="step", method_name="to_step_stream", text=True),
                     ExportFormat(selector="stl", extension="stl", method_name="to_stl_stream", text=False),
                     ExportFormat(selector="html", extension="html", method_name="to_html_stream", text=True)]
        formats.extend(formats3d)
        return formats


class MovingObject(PhysicalObject):
    """ A 3D object which display can move down a path from according to defined steps. """

    def volmdlr_primitives_step_frames(self):
        """ Return a list of volmdlr primitives to build up volume model. """
        raise NotImplementedError('Object inheriting MovingObject should implement volmdlr_primitives_step_frames')

    def volmdlr_volume_model(self, **kwargs):
        """ Volume model of Moving Object. """
        import volmdlr as vm  # !!! Avoid circular imports, is this OK ?
        return vm.core.MovingVolumeModel(self.volmdlr_primitives(**kwargs), self.volmdlr_primitives_step_frames())


class Parameter(DessiaObject):
    """ A value from a Parameter Set. """

    def __init__(self, lower_bound, upper_bound, periodicity=None, name=''):
        DessiaObject.__init__(self, name=name)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.periodicity = periodicity

    def random_value(self):
        """ Sample a value within the bounds. """
        return random.uniform(self.lower_bound, self.upper_bound)

    def are_values_equal(self, value1, value2, tol=1e-2):
        """ Check equality according to given tolerance. """
        if self.periodicity is not None:
            value1 = value1 % self.periodicity
            value2 = value2 % self.periodicity

        return math.isclose(value1, value2, abs_tol=tol)

    def normalize(self, value):
        """ Translate the value to a [0, 1] interval where 0 are the lower and upper bounds. """
        normalized_value = (value - self.lower_bound) / (self.upper_bound - self.lower_bound)
        return normalized_value

    def original_value(self, normalized_value):
        """ Revert normalized value to its original value. """
        value = normalized_value * (self.upper_bound - self.lower_bound) + self.lower_bound
        return value

    def optimizer_bounds(self):
        """ Compute optimize bounds. """
        if self.periodicity is not None:
            return self.lower_bound - 0.5 * self.periodicity, self.upper_bound + 0.5 * self.periodicity
        return None


class ParameterSet(DessiaObject):
    """ Object that can provide miscellaneous features around values Dataset. """

    def __init__(self, values, name=''):
        self.values = values

        DessiaObject.__init__(self, name=name)

    @property
    def parameters(self):
        """ Compute parameters from ParameterSet values. """
        parameters = [Parameter(min(v), max(v), name=k) for k, v in self.values.items()]
        return parameters

    @property
    def means(self):
        """ Compute means on all parameters defined in Set. """
        means = {k: sum(v) / len(v) for k, v in self.values.items()}
        return means


class DessiaFilter(DessiaObject):
    """
    Base class for filters working on lists of DessiaObjects (List[DessiaObject]).

    :param attribute: Name of attribute on which to filter
    :type attribute: str

    :param comparison_operator: Comparison operator
    :type comparison_operator: str

    :param bound: The bound value to compare `'attribute'` of DessiaObjects of a list with `'comparison_operator'`
    :type bound: float

    :param name: Name of filter
    :type name: `str`, `optional`, defaults to `''`

    :Comparison operators:
        * greater than: >=, gte, ge
        * greater: >, gt
        * lower than: <=, lte, le
        * lower: <, lt
        * lower: <, lt
        * equal: ==, eq
        * different: !=, ne
    """

    _REAL_OPERATORS = {'>': operator.gt, '<': operator.lt, '>=': operator.ge, '<=': operator.le, '==': operator.eq,
                       '!=': operator.ne, 'gt': operator.gt, 'lt': operator.lt, 'ge': operator.ge, 'le': operator.le,
                       'eq': operator.eq, 'ne': operator.ne, 'gte': operator.ge, 'lte': operator.le}

    def __init__(self, attribute: str, comparison_operator: str, bound: float, name: str = ''):
        self.attribute = attribute
        self.comparison_operator = comparison_operator
        self.bound = bound
        DessiaObject.__init__(self, name=name)

    def __str__(self, offset_attr: int = 10, offset_boun: int = 0):
        offset_oper = 0
        if offset_boun == 0:
            offset_boun = len(str(self.bound)) + 2
        string_operator = {'>': '>', '<': '<', '>=': '>=', '<=': '<=', '==': '==', '!=': '!=', 'gt': '>', 'lt': '<',
                           'ge': '>=', 'le': '<=', 'eq': '==', 'ne': '!=', 'gte': '>=', 'lte': '<='}
        printed_operator = string_operator[self.comparison_operator]
        return (self.attribute + " " * (offset_attr - len(self.attribute)) +
                printed_operator + " " * (offset_oper - len(printed_operator)) +
                " " * (offset_boun - len(str(self.bound))) + str(self.bound))

    def __hash__(self):
        hash_ = len(self.attribute)
        hash_ += hash(self.comparison_operator)
        hash_ += hash(self.bound)
        return int(hash_)

    def __eq__(self, other: 'DessiaFilter'):
        same_attr = self.attribute == other.attribute
        same_op = self.comparison_operator == other.comparison_operator
        same_bound = self.bound == other.bound
        return same_attr and same_op and same_bound

    def _comparison_operator(self):
        return self._REAL_OPERATORS[self.comparison_operator]

    def _to_lambda(self):
        return lambda x: (self._comparison_operator()(get_in_object_from_path(value, f'#/{self.attribute}'),
                                                      self.bound) for value in x)

    def get_booleans_index(self, values: List[DessiaObject]):
        """
        Get the Boolean indexing of a filtered list.

        :param values: List of DessiaObjects to filter
        :type values: List[DessiaObject]

        :return: `list of length `len(values)` where the elements are `True` if kept by the filter, otherwise `False`.
        :rtype: List[bool]

        :Examples:
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> filter_.get_booleans_index(values)
        [False, False, True, True, True]
        """
        return list(self._to_lambda()(values))

    @staticmethod
    def booleanlist_to_indexlist(booleans_list: List[int]):  # TODO: Should it exist ?
        """
        Transform a Boolean list to an index list.

        :param booleans_list: list of length `len(values)` where elements are `True` if kept, otherwise `False`.
        :type booleans_list: List[int]

        :return: list of kept indexes
        :rtype: List[int]

        :Examples:
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> booleans_list = filter_.get_booleans_index(values)
        [False, False, True, True, True]
        >>> DessiaFilter.booleanlist_to_indexlist(booleans_list)
        [2, 3, 4]
        """
        return list(itertools.compress(range(len(booleans_list)), booleans_list))

    @staticmethod
    def apply(values: List[DessiaObject], booleans_list: List[List[bool]]):
        """
        Apply a Dessia Filter on a list of DessiaObjects.

        :param values: List of DessiaObjects to filter
        :type values: List[DessiaObject]

        :param booleans_list: list of length `len(values)` where elements are `True` if kept, otherwise `False`.
        :type booleans_list: List[List[bool]]

        :return: List of filtered values
        :rtype: List[DessiaObject]

        :Examples:
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> booleans_list = filter_.get_booleans_index(values)
        [False, False, True, True, True]
        >>> for car in DessiaFilter.apply(values, booleans_list): print(car.weight)
        3436.0, 3433.0, 3449.0
        """
        return list(itertools.compress(values, booleans_list))


class FiltersList(DessiaObject):
    """
    Combine several filters stored as a list of DessiaFilters with a logical operator.

    :param filters: List of DessiaFilters to combine
    :type filters: List[DessiaFilter]

    :param logical_operator: Logical operator to combine filters
    :type logical_operator: str

    :param name: Name of FiltersList
    :type name: `str`, `optional`, defaults to `''`

    :Logical operators: `'and'`, `'or'`, `'xor'`
    """

    _standalone_in_db = True

    def __init__(self, filters: List[DessiaFilter], logical_operator: str = 'and', name: str = ''):
        self.filters = filters
        self.logical_operator = logical_operator
        DessiaObject.__init__(self, name=name)

    def __len__(self):
        return len(self.filters)

    def __str__(self):
        print_lim = 15
        len_attr = max(map(len, [filter_.attribute for filter_ in self.filters]))
        len_numb = max(map(len, [str(filter_.bound) for filter_ in self.filters]))
        prefix = f"{self.__class__.__name__} {self.name if self.name != '' else hex(id(self))}: "
        prefix += f"{len(self)} filters combined with '" + self.logical_operator + "' operator :\n"
        string = ""
        for filter_ in self.filters[:print_lim]:
            string += " " * 3 + "- "
            string += filter_.__str__(len_attr + 2, len_numb + 2)
            string += "\n"
        return prefix + string

    @classmethod
    def from_filters_list(cls, filters: List[DessiaFilter], logical_operator: str = 'and', name: str = ''):
        """
        Compute a FilersList from an already built list of DessiaFilter.

        :param filters: List of DessiaFilters to combine
        :type filters: List[DessiaFilter]

        :param logical_operator: Logical operator to combine filters (`'and'`, `'or'` or `'xor'`)
        :type logical_operator: `str`, `optional`, defaults to `'and'`

        :param name: Name of FiltersList
        :type name: `str`, `optional`, defaults to `''`

        :return: A new instantiated list of DessiaFilter
        :rtype: FiltersList

        :Examples:
        >>> from dessia_common.core import DessiaFilter, FiltersList
        >>> filters = [DessiaFilter('weight', '<=', 3500.), DessiaFilter('mpg', '<=', 40.)]
        >>> filters_list = FiltersList(filters, logical_operator="or", name="example")
        >>> print(filters_list)
        FiltersList example: 2 filters combined with 'or' operator :
           - weight  <=  3500.0
           - mpg     <=    40.0
        """
        return cls(filters=filters, logical_operator=logical_operator, name=name)

    @staticmethod
    def combine_booleans_lists(booleans_lists: List[List[bool]], logical_operator: str = "and") -> List[bool]:
        """
        Combine a list of Boolean indices with the logical operator into a simple Boolean index.

        :param booleans_lists: List of boolean indices

        :param logical_operator: Logical operator to combine filters (`'or'`, `'and'` or `'xor'`)

        :raises NotImplementedError: If logical_operator is not one of `'and'`, `'or'`, `'xor'`, raises an error

        :return: Boolean indices of the filtered data

        :Examples:
        >>> from dessia_common.core import FiltersList
        >>> booleans_lists = [[True, True, False, False], [False, True, True, False]]
        >>> FiltersList.combine_booleans_lists(booleans_lists, logical_operator="xor")
        [True, False, True, False]
        """
        if logical_operator == 'and':
            return [all(booleans_tuple) for booleans_tuple in zip(*booleans_lists)]
        if logical_operator == 'or':
            return [any(booleans_tuple) for booleans_tuple in zip(*booleans_lists)]
        if logical_operator == 'xor':
            return [bool(sum(booleans_tuple)) for booleans_tuple in zip(*booleans_lists)]
        raise NotImplementedError(f"'{logical_operator}' str for 'logical_operator' attribute is not a use case")

    def get_booleans_index(self, dobjects_list: List[DessiaObject]):
        """
        Compute all the filters of `self.filters` on `dobjects_list` and return boolean indices of `dobjects_list`.

        :param dobject_list: List of data to filter
        :type dobject_list: List[DessiaObject]

        :return: A `booleans index` of `dobjects_list` of the list of data to filter (`dobjects_list`)
        :rtype: List[bool]

        :Examples:
        >>> from dessia_common.core import FiltersList
        >>> from dessia_common.models import all_cars_no_feat
        >>> dobjects_list = all_cars_no_feat[:5]
        >>> filters = [DessiaFilter('weight', '<=', 4000.), DessiaFilter('mpg', '>=', 30.)]
        >>> filters_list = FiltersList(filters, logical_operator="xor", name="example")
        >>> filters_list.get_booleans_index(dobjects_list)
        [True, True, True, True, True]
        """
        booleans_index = []
        for filter_ in self.filters:
            booleans_index.append(filter_.get_booleans_index(dobjects_list))
        return self.__class__.combine_booleans_lists(booleans_index, self.logical_operator)

    def apply(self, dobjects_list: List[DessiaObject]):
        """
        Apply a FiltersList on a list of DessiaObjects.

        :param dobjects_list: List of DessiaObjects to filter
        :type dobjects_list: List[DessiaObject]

        :return: List of filtered values
        :rtype: List[DessiaObject]

        :Examples:
        >>> from dessia_common.core import FiltersList
        >>> from dessia_common.datatools.dataset import Dataset
        >>> from dessia_common.models import all_cars_wi_feat
        >>> filters = [DessiaFilter('weight', '<=', 1650.), DessiaFilter('mpg', '>=', 45.)]
        >>> filters_list = FiltersList(filters, logical_operator="xor", name="example")
        >>> filtered_cars = filters_list.apply(all_cars_wi_feat)
        >>> print(Dataset(filtered_cars, name="example"))
        Dataset example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               35.0  |             0.072  |              69.0  |            1613.0  |              18.0  |
        |               31.0  |             0.076  |              52.0  |            1649.0  |              16.5  |
        |               46.6  |             0.086  |              65.0  |            2110.0  |              17.9  |
        """
        booleans_index = self.get_booleans_index(dobjects_list)
        return DessiaFilter.apply(dobjects_list, booleans_index)


def stringify_dict_keys(obj):
    """ Stringify dict keys. """
    if isinstance(obj, (list, tuple)):
        new_obj = []
        for elt in obj:
            new_obj.append(stringify_dict_keys(elt))

    elif isinstance(obj, dict):
        new_obj = {}
        for key, value in obj.items():
            new_obj[str(key)] = stringify_dict_keys(value)
    else:
        return obj
    return new_obj


def type_from_annotation(type_, module):
    """ Clean up a proposed type if there are stringified. """
    if isinstance(type_, str):
        # Evaluating types
        type_ = dcs.TYPES_FROM_STRING.get(type_, default=getattr(import_module(module), type_))
    return type_


def prettyname(namestr):
    """ Create a pretty name from as str. """
    warnings.warn("prettyname function has been moved to 'helpers' module. Use it instead", DeprecationWarning)
    return dch.prettyname(namestr)


def inspect_arguments(method, merge=False):
    """
    Find default value and required arguments of class construction.

    Get method arguments and default arguments as sequences while removing forbidden ones (self, cls...).
    """
    warnings.warn("Method 'inspect_arguments' have been moved to dessia_common/schemas."
                  "Use it instead instead", DeprecationWarning)
    return dcs.inspect_arguments(method=method, merge=merge)


def split_default_args(argspecs, merge: bool = False):
    """
    Find default value and required arguments of class construction.

    Get method arguments and default arguments as sequences while removing forbidden ones (self, cls...).
    """
    warnings.warn("Method 'split_default_args' have been moved to dessia_common/schemas."
                  "Use it instead instead", DeprecationWarning)
    return dcs.split_default_args(argspecs=argspecs, merge=merge)


def split_argspecs(argspecs) -> Tuple[int, int]:
    """ Get number of regular arguments as well as arguments with default values. """
    warnings.warn("Method 'split_argspecs' have been moved to dessia_common/schemas."
                  "Use it instead instead", DeprecationWarning)
    return dcs.split_argspecs(argspecs=argspecs)


def get_attribute_names(object_class):
    """ Get all attributes of a class which are present in __init__ or numeric attributes and not in parent class. """
    attributes = [attribute[0] for attribute in inspect.getmembers(object_class, lambda x: not inspect.isroutine(x))
                  if not attribute[0].startswith('__')
                  and not attribute[0].endswith('__')
                  and isinstance(attribute[1], (float, int, complex, bool))]
    subclass_attributes = {name: param for name, param in inspect.signature(object_class.__init__).parameters.items()
                           if type in inspect.getmro(param.annotation.__class__)}
    subclass_numeric_attributes = [name for name, param in subclass_attributes.items()
                                   if any(item in inspect.getmro(param.annotation)
                                          for item in [float, int, bool, complex])]
    attributes += [a for a in subclass_numeric_attributes if a not in dcs.RESERVED_ARGNAMES]
    return attributes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A simple workflow composed of functions
"""
import json
import dessia_common.typings as dct
from dessia_common.workflow.core import Pipe, Workflow, WorkflowRun
from dessia_common.workflow.blocks import InstantiateModel, ModelMethod, ModelAttribute, WorkflowBlock, ForEach,\
    Unpacker, Sequence, Filter
from dessia_common.core import DessiaFilter

import dessia_common.tests as dctests

instanciate_generator = InstantiateModel(model_class=dctests.Generator, name='Instantiate Generator')
generator_generate = ModelMethod(dct.MethodType(dctests.Generator, 'generate'), name='Generator Generate')
attribute_selection = ModelAttribute(attribute_name='models', name='Attribute Selection')

# Subworkflow of model optimization
instanciate_optimizer = InstantiateModel(model_class=dctests.Optimizer, name='Instantiate Optimizer')
optimization = ModelMethod(dct.MethodType(dctests.Optimizer, 'optimize'), name='Optimization')
model_fetcher = ModelAttribute(attribute_name='model_to_optimize', name='Model Fetcher')

pipe1_opt = Pipe(input_variable=instanciate_optimizer.outputs[0], output_variable=optimization.inputs[0])
pipe2_opt = Pipe(input_variable=optimization.outputs[1], output_variable=model_fetcher.inputs[0])
optimization_blocks = [instanciate_optimizer, optimization, model_fetcher]
optimization_pipes = [pipe1_opt, pipe2_opt]
optimization_workflow = Workflow(blocks=optimization_blocks, pipes=optimization_pipes,
                                 output=model_fetcher.outputs[0], name='Optimization Workflow')

optimization_workflow_block = WorkflowBlock(workflow=optimization_workflow, name='Workflow Block')

parallel_optimization = ForEach(workflow_block=optimization_workflow_block, iter_input_index=0, name='ForEach')

unpacker = Unpacker(indices=[0, 3, -1], name='Unpacker')
sequence = Sequence(number_arguments=2, name='Sequence')

filters = [DessiaFilter(attribute='value', comparison_operator='gt', bound=0),
           DessiaFilter(attribute='submodel/subvalue', comparison_operator='lt', bound=2000)]

filter_sort = Filter(filters=filters, name='Filters')

pipe_1 = Pipe(input_variable=instanciate_generator.outputs[0], output_variable=generator_generate.inputs[0])
pipe_2 = Pipe(input_variable=generator_generate.outputs[1], output_variable=attribute_selection.inputs[0])
pipe_3 = Pipe(input_variable=attribute_selection.outputs[0], output_variable=parallel_optimization.inputs[0])
pipe_4 = Pipe(input_variable=parallel_optimization.outputs[0], output_variable=unpacker.inputs[0])
pipe_51 = Pipe(input_variable=unpacker.outputs[0], output_variable=sequence.inputs[0])
pipe_52 = Pipe(input_variable=unpacker.outputs[2], output_variable=sequence.inputs[1])
pipe_6 = Pipe(input_variable=sequence.outputs[0], output_variable=filter_sort.inputs[0])

blocks = [instanciate_generator, generator_generate, attribute_selection,
          parallel_optimization, unpacker, sequence, filter_sort]
pipes = [pipe_1, pipe_2, pipe_3, pipe_4, pipe_51, pipe_52, pipe_6]
demo_workflow = Workflow(blocks=blocks, pipes=pipes, output=unpacker.outputs[0])

input_values = {0: 5}

demo_workflow_run = demo_workflow.run(input_values=input_values, verbose=True)

# Assert to_dict, dict_to_object, hashes, eqs
dict_ = demo_workflow_run.to_dict()
demo_workflow_run2 = WorkflowRun.dict_to_object(dict_=dict_)

assert hash(demo_workflow_run) == hash(demo_workflow_run2)

assert demo_workflow_run2 == demo_workflow_run


demo_workflow_run_copy = demo_workflow_run.copy()
assert demo_workflow_run == demo_workflow_run_copy

demo_workflow._check_platform()
demo_workflow_run._check_platform()

# Assert deserialization
demo_workflow_dict = demo_workflow.to_dict()

demo_workflow_json = json.dumps(demo_workflow_dict)
dict_from_json = json.loads(demo_workflow_json)
deserialized_demo_workflow = Workflow.dict_to_object(dict_from_json)
assert demo_workflow == deserialized_demo_workflow

# Worflow State
demo_workflow_state = demo_workflow.start_run({})
demo_workflow_state.add_input_value(0, 5)
demo_workflow_state.evaluate_next_block()

assert len(demo_workflow_state.values) == 1

copy_workflow_state = demo_workflow_state.copy()

assert copy_workflow_state == demo_workflow_state

demo_workflow_state._check_platform()

assert len(demo_workflow_run.values) == 0  # It has no export/display, so values should not be stored

print("workflow_with_models.py has passed")



""" Script for workflow with exports creationZ. """

from dessia_common.workflow.core import Variable, Pipe, Workflow
from dessia_common.workflow.blocks import InstantiateModel, ModelMethod, ModelAttribute, WorkflowBlock, ForEach,\
    Export, Unpacker, Archive, MultiPlot
from dessia_common.forms import Generator, Optimizer, StandaloneObject
from dessia_common.typings import MethodType

inst = InstantiateModel(model_class=Generator, name='Instantiate Generator')

generate_method = MethodType(class_=Generator, name='generate')
gene = ModelMethod(method_type=generate_method, name='Generator Generate')
attr = ModelAttribute(attribute_name='models', name='Attribute Selection')

# Sub Workflow of model optimization
opti = InstantiateModel(model_class=Optimizer, name='Instantiate Optimizer')


optimize_method = MethodType(class_=Optimizer, name='optimize')
optimization = ModelMethod(method_type=optimize_method, name='Optimization')
model_fetcher = ModelAttribute(attribute_name='model_to_optimize', name='Model Fetcher')

pipe1_opt = Pipe(input_variable=opti.outputs[0], output_variable=optimization.inputs[0])
pipe2_opt = Pipe(input_variable=optimization.outputs[1], output_variable=model_fetcher.inputs[0])
opti_blocks = [opti, optimization, model_fetcher]
opti_pipes = [pipe1_opt, pipe2_opt]
opti_workflow = Workflow(blocks=opti_blocks, pipes=opti_pipes,
                         output=model_fetcher.outputs[0], name='Optimization Workflow')

optimization_workflow_block = WorkflowBlock(workflow=opti_workflow, name='Workflow Block')

parallel_optimization = ForEach(workflow_block=optimization_workflow_block, iter_input_index=0, name='ForEach')

display_attributes = ['intarg', 'strarg', 'standalone_subobject/floatarg']
display_ = MultiPlot(selector_name="Multiplot", attributes=display_attributes, name='Display')

unpack_results = Unpacker(indices=[0], name="Unpack Results")

txt_method = MethodType(class_=Generator, name="save_to_stream")
export_txt = Export(method_type=txt_method, text=True, filename="export_json", extension="json", name="Export JSON")
xlsx_method = MethodType(class_=StandaloneObject, name="to_xlsx_stream")
export_xlsx = Export(method_type=xlsx_method, text=False, filename="export_xlsx",
                     extension="xlsx", name="Export XLSX")

zip_export = Archive(number_exports=2, name="Zip")

int_variable = Variable(type_=int, name="Some Integer")

pipe_int_1 = Pipe(input_variable=int_variable, output_variable=inst.inputs[1])
pipe_1 = Pipe(input_variable=inst.outputs[0], output_variable=gene.inputs[0])
pipe_2 = Pipe(input_variable=gene.outputs[1], output_variable=attr.inputs[0])
pipe_3 = Pipe(input_variable=attr.outputs[0], output_variable=parallel_optimization.inputs[0])
pipe_4 = Pipe(input_variable=parallel_optimization.outputs[0], output_variable=unpack_results.inputs[0])
pipe_5 = Pipe(input_variable=inst.outputs[0], output_variable=export_txt.inputs[0])
pipe_6 = Pipe(input_variable=unpack_results.outputs[0], output_variable=export_xlsx.inputs[0])

pipe_display = Pipe(input_variable=parallel_optimization.outputs[0], output_variable=display_.inputs[0])

pipe_export_1 = Pipe(input_variable=export_txt.outputs[0], output_variable=zip_export.inputs[0])
pipe_export_2 = Pipe(input_variable=export_xlsx.outputs[0], output_variable=zip_export.inputs[1])

blocks = [inst, gene, attr, parallel_optimization,
          display_, unpack_results, export_txt, zip_export, export_xlsx]
pipes = [pipe_int_1, pipe_1, pipe_2, pipe_3, pipe_4, pipe_5, pipe_6, pipe_display, pipe_export_1, pipe_export_2]
workflow_export = Workflow(blocks=blocks, pipes=pipes, output=parallel_optimization.outputs[0],
                           name="Workflow Test Export")

# workflow_export.save_script_to_file('workflow_export')

workflow_export_state = workflow_export.start_run()
workflow_export_state.name = "WorkflowState Test Export"

input_values = {workflow_export.input_index(inst.inputs[0]): 0,
                workflow_export.input_index(inst.inputs[1]): 5}

workflow_export_run = workflow_export.run(input_values)



from dessia_common.core import DessiaObject
from dessia_common.utils.diff import diff

value1 = {"a": 1, "b": 2, "c": 3}
value2 = {"a": 1, "b": 2}
value3 = {"a": 1}

# Checking that full path are given in result
assert diff(value1=value1, value2=value2).__repr__() == "Missing attributes:\t\t#/c: missing in second object\n"
assert diff(value1=value1, value2=value3).__repr__() == "Missing attributes:\t\t#/b: missing in second object\n\t" \
                                                        "#/c: missing in second object\n"

# Checking that diff is robust to commutation
assert diff(value1=value2, value2=value3).__repr__() == "Missing attributes:\t\t#/b: missing in second object\n"
assert diff(value1=value3, value2=value2).__repr__() == "Missing attributes:\t\t#/b: missing in first object\n"


class ObjTestEq(DessiaObject):
    """ Test Simple equalities. """
    _non_data_eq_attributes = ['arg3']
    _non_data_hash_attributes = ['arg3']

    def __init__(self, arg1: int, arg2: int):
        self.arg1 = arg1
        self.arg2 = arg2
        self.arg3 = 0

        DessiaObject.__init__(self, name="")


obj_1 = ObjTestEq(1, 2)
obj_2 = ObjTestEq(1, 2)
assert obj_2 == obj_1
obj_1.arg3 = 5
assert obj_2 == obj_1

print("script diff.py has passed")



""" Displays for dessia_common. """
import webbrowser
import os
import tempfile
import inspect
import json


from typing import Union, Optional
from networkx import DiGraph, Graph, kamada_kawai_layout
from dessia_common.templates import visjs_template
from dessia_common.typings import JsonSerializable


class DisplaySetting:
    """ Describe which method to call to get a display. """

    def __init__(self, selector: Optional[str], type_: str, method: str, arguments=None,
                 serialize_data: bool = False, load_by_default: bool = False):
        self.selector = selector
        self.type = type_
        self.method = method
        if not arguments:
            arguments = {}
        self.arguments = arguments
        self.serialize_data = serialize_data
        self.load_by_default = load_by_default

    @property
    def reference_path(self) -> str:
        """
        Get reference path or set it to its default value if not defined.

        Experimental.
        """
        if self.arguments is None:
            return "#"

        reference_path = self.arguments.get("reference_path", "#")
        if reference_path is None:
            return "#"
        return reference_path

    def to_dict(self):
        """ Serialization: make a dict from class instance attributes. """
        return {"selector": self.selector, "type": self.type, "method": self.method, "arguments": self.arguments,
                "serialize_data": self.serialize_data, "load_by_default": self.load_by_default}

    def compose(self, attribute: str):
        """
        Handles deep calls to method.

        In case of a parent getting the display settings of a children, this method allow
        to inject the attribute name to method name.
        """
        self.method = f"{attribute}.{self.method}"


class DisplayObject:
    """ Container for data of display. A traceback can be set if display fails to be generated. """

    def __init__(self, type_: str, data: Union[JsonSerializable, str],
                 reference_path: str = '', traceback: str = '', name: str = ''):
        self.type_ = type_
        self.data = data
        self.traceback = traceback
        self.reference_path = reference_path
        self.name = name

        if data and type_ == "markdown":
            self.data_cleaning()

    def data_cleaning(self):
        """ Cleanup tabs in markdown. """
        self.data = inspect.cleandoc(self.data)

    def to_dict(self):
        """ Simple serialization. """
        return {"type_": self.type_, "data": self.data, "traceback": self.traceback,
                "reference_path": self.reference_path, "name": self.name}


def networkx_to_visjs_data(networkx_graph: Graph):
    """ Compute visjs data to plot from a networkx graph. """
    visjs_data = {"name": networkx_graph.name, "nodes": [], "edges": []}

    pos = kamada_kawai_layout(networkx_graph)

    for i, node in enumerate(networkx_graph.nodes):
        node_dict = networkx_graph.nodes[node]
        node_data = {"id": i}

        if node in pos:
            node_data["x"], node_data["y"] = pos[node]

        if "name" not in node_dict and "label" not in node_dict:
            if isinstance(node, str):
                node_data["label"] = node
            elif isinstance(node, int):
                node_data["label"] = str(node)
            else:
                node_data["label"] = ''
            node_data["title"] = node_data["label"]
            if len(node_data["label"]) > 10:
                node_data["label"] = "[ ]"
        elif "name" in node_dict and "label" not in node_dict:
            node_data["label"] = node_dict["name"]
        else:
            node_data["label"] = node_dict["name"] + node_dict["label"]

        if "shape" not in node_dict:
            node_data["shape"] = "circular"
        else:
            node_data["shape"] = node_dict["shape"]

        if "color" in node_dict:
            node_data["color"] = node_dict["color"]
        visjs_data["nodes"].append(node_data)

    list_nodes = list(networkx_graph.nodes)
    is_digraph = isinstance(networkx_graph, DiGraph)
    for edge in networkx_graph.edges:
        index1 = list_nodes.index(edge[0])
        index2 = list_nodes.index(edge[1])
        edge_nx_data = networkx_graph.get_edge_data(*edge)
        edge_data = {"from": index1, "to": index2, "font": {"align": "middle"}}
        if is_digraph:
            if "head_type" in edge_nx_data:
                edge_data["arrows"] = {"to": {"enabled": True, "type": edge_nx_data["head_type"]}}
            else:
                edge_data["arrows"] = "to"
        if "color" in edge_nx_data:
            edge_data["color"] = {"color": edge_nx_data["color"]}
        visjs_data["edges"].append(edge_data)
    return visjs_data


def draw_networkx_graph(networkx_graph: Graph):
    """ Draw a networkx graph in a browser using VisJS library. """
    visjs_data = networkx_to_visjs_data(networkx_graph)
    content = visjs_template.substitute(nodes=json.dumps(visjs_data["nodes"]),
                                        edges=json.dumps(visjs_data["edges"]),
                                        name=visjs_data["name"])
    with tempfile.NamedTemporaryFile(suffix=".html", delete=False) as file:
        file.write(bytes(content, "utf8"))
    webbrowser.open("file://" + os.path.realpath(file.name))
    return file.name



"""
Cluster.py package testing.
"""
import json
from dessia_common.models import all_cars_no_feat, all_cars_wi_feat, rand_data_small, rand_data_large
from dessia_common.datatools import HeterogeneousList, CategorizedList
from dessia_common.datatools.dataset import Dataset
from dessia_common.datatools.cluster import ClusteredDataset

# When attribute _features is not specified in class Car
all_cars_without_features = Dataset(all_cars_no_feat)
# When attribute _features is specified in class CarWithFeatures
all_cars_with_features = Dataset(all_cars_wi_feat)
# Auto-generated heterogeneous small dataset with nb_clusters clusters of points in nb_dims dimensions
small_RandDatas_heterogeneous = Dataset(rand_data_small)
# Auto-generated heterogeneous large dataset with nb_clusters clusters of points in nb_dims dimensions
big_RandDatas_heterogeneous = Dataset(rand_data_large)

# Build ClusteredDatasets
clustered_cars_without = ClusteredDataset.from_dbscan(all_cars_without_features, eps=40)
clustered_cars_with = ClusteredDataset.from_dbscan(all_cars_with_features, eps=40)
aggclustest_clustered = ClusteredDataset.from_agglomerative_clustering(big_RandDatas_heterogeneous, n_clusters=10)
kmeanstest_clustered = ClusteredDataset.from_kmeans(small_RandDatas_heterogeneous, n_clusters=10, scaling=True)

clustered_cars_with_list = ClusteredDataset.list_dbscan(all_cars_no_feat, eps=40)
aggclustest_clustered_list = ClusteredDataset.list_agglomerative_clustering(rand_data_large, n_clusters=10)
kmeanstest_clustered_list = ClusteredDataset.list_kmeans(rand_data_small, n_clusters=10, scaling=True)
assert(clustered_cars_with_list == clustered_cars_without)
assert(aggclustest_clustered_list == aggclustest_clustered)
assert(kmeanstest_clustered_list.n_clusters == 10)


# Split lists into labelled lists
split_cars_without = clustered_cars_without.clustered_sublists()
split_cars_with = clustered_cars_with.clustered_sublists()
aggclustest_split = aggclustest_clustered.clustered_sublists()
kmeanstest_split = kmeanstest_clustered.clustered_sublists()

# Centroids
assert(clustered_cars_with.cluster_real_centroids('mahalanobis')[0].to_vector()[1] == 0.135)
assert(clustered_cars_with.cluster_real_centroids('minkowski')[0].to_vector()[1] == 0.156)

# Test print
clustered_cars_without.labels[0] = 15000
clustered_cars_without.labels[1] = -1
clustered_cars_without.labels[2:100] = [999999] * len(clustered_cars_without[2:100])
print(clustered_cars_without)
hlist = Dataset(all_cars_wi_feat, name="cars")
clist = ClusteredDataset.from_agglomerative_clustering(hlist, n_clusters=10, name="cars")
split_clist = clist.clustered_sublists()
split_clist[0].name = "15g6e4rg84reh56rt4h56j458hrt56gb41rth674r68jr6"
print(split_clist)

# Test ClusterResults instances on platform
clustered_cars_without._check_platform()
clustered_cars_with._check_platform()
aggclustest_clustered._check_platform()
kmeanstest_clustered._check_platform()

# Test plots outside platform
cluscars_plot_data = clustered_cars_without.plot_data()
cluscars_plot_data = split_cars_with.plot_data()
# assert(json.dumps(cluscars_plot_data[0].to_dict())[42500:42550] == ' 2515.0, "acceleration": 14.8, "model": 78.0, "Clu')
# assert(json.dumps(cluscars_plot_data[1].to_dict())[52500:52550] == '4.3, "model": 80.0, "Cluster Label": -1}, {"mpg": ')
# assert(json.dumps(cluscars_plot_data[2].to_dict())[50:100] == 'te_names": ["Index of reduced basis vector", "Sing')


# =============================================================================
# JSON TESTS
# =============================================================================
dict_cars_without = clustered_cars_without.to_dict(use_pointers=True)
dict_cars_with = clustered_cars_with.to_dict(use_pointers=True)
dict_aggclustest = aggclustest_clustered.to_dict(use_pointers=True)
dict_kmeanstest = kmeanstest_clustered.to_dict(use_pointers=True)

# Cars without features
json_dict = json.dumps(dict_cars_without)
decoded_json = json.loads(json_dict)
deserialized_object = clustered_cars_without.dict_to_object(decoded_json)

# Cars with features
json_dict = json.dumps(dict_cars_with)
decoded_json = json.loads(json_dict)
deserialized_object = clustered_cars_with.dict_to_object(decoded_json)

# Small dataset
json_dict = json.dumps(dict_aggclustest)
decoded_json = json.loads(json_dict)
deserialized_object = aggclustest_clustered.dict_to_object(decoded_json)

# Large dataset
json_dict = json.dumps(dict_kmeanstest)
decoded_json = json.loads(json_dict)
deserialized_object = kmeanstest_clustered.dict_to_object(decoded_json)

# Missing tests after coverage report
try:
    clustered_cars_without + clustered_cars_without
    raise ValueError("ClusteredDataset should be summable")
except Exception as e:
    assert(e.args[0] == "Addition only defined for Dataset. A specific __add__ method is required for " +
           "<class 'dessia_common.datatools.cluster.ClusteredDataset'>")

# Exports XLS
clustered_cars_without.to_xlsx('clus_xls.xlsx')
split_cars_with.to_xlsx('clus_xls_2.xlsx')

# Retrocompatibility
Hlist = HeterogeneousList(all_cars_no_feat)
Clist = CategorizedList(all_cars_no_feat)
